@@ -47,6 +47,7 @@ import copy
 import itertools
 import sys
 
+import six
 from zope.component import getUtility, getGlobalSiteManager
 from zope.interface import classImplements
 from zope.interface.advice import addClassAdvisor
@@ -405,7 +406,7 @@ def exported(field, *versioned_annotations, **kwparams):
     # If any keywords are left over, raise an exception.
     if len(kwparams) > 0:
         raise TypeError("exported got an unexpected keyword "
-                        "argument '%s'" % kwparams.keys()[0])
+                        "argument '%s'" % list(kwparams)[0])
 
     # Now incorporate the list of named dicts into the VersionedDict.
     for version, annotations in reversed(versioned_annotations):
@@ -738,7 +739,8 @@ def _update_default_and_required_params(params, method_info):
             # zope schema are expected to be unicode, whereas it's
             # really possible that the method's default is a simple
             # string.
-            if isinstance(default, str) and IText.providedBy(param_def):
+            if (six.PY2 and
+                    isinstance(default, str) and IText.providedBy(param_def)):
                 default = unicode(default)
             param_def.default = default
             param_def.required = False
@@ -1042,9 +1044,9 @@ class cache_for(_method_annotator):
 
     def __init__(self, duration):
         """Specify the duration, in seconds, of the caching resource."""
-        if not isinstance(duration, (int, long)):
+        if not isinstance(duration, six.integer_types):
             raise TypeError(
-                'Caching duration should be int or long, not %s' %
+                'Caching duration should be an integer type, not %s' %
                 duration.__class__.__name__)
         if duration <= 0:
             raise ValueError(
@@ -1254,7 +1256,7 @@ def generate_entry_adapters(
     # Go through the fields and build up a picture of what this entry looks
     # like for every version.
     adapters_by_version = {}
-    fields = getFields(content_interface).items()
+    fields = list(getFields(content_interface).items())
     for version, iface in webservice_interfaces:
         fields.extend(getFields(iface).items())
     for name, field in fields:
@@ -1805,5 +1807,5 @@ def _versioned_class_name(base_name, version):
         # runtime. Use a generic string that won't conflict with a
         # real version string.
         version = "__Earliest"
-    name = "%s_%s" % (base_name, version.encode('utf8'))
+    name = "%s_%s" % (base_name, six.ensure_str(version))
     return make_identifier_safe(name)
